%Image and Vision Computing (2024) https://doi.org/10.1016/j.imavis.2024.105192
%by Na Lei, Jisui Huang, Ke Chen, Yuxue Ren, Emil Saucan, Zhenchang Wang and Yuanyuan Shang  
%
%Notation:
% Img: the input 3D image, e.g., of size [1, 256, 256, 64], or [4, 256, 256, 64]
% Lev_ini: see "get_image" for markers' points to roughly locate the object
%
%Notation for output:  seg (binary) and u
%(1) Input image = Img:   of size Img=[1, 256, 256, 64], or [d, 256, 256, 64]
%        where d denotes the number of image modalities given 
%                  e.g. d=2 if bpth T1 and T2 scans are given
%(2) Input Lvset_ini: Initialised label Lv=[256,256,64] (geometric constraint)
%(3) Label_gt: Ground label (seg) for comparison use only (if not, set it to Img)
clear; close all; eval(['help ' mfilename])
addpath('.\c++\level_set\build\Release')
addpath('.\c++\ricci\build\Release\') % path="BraTS2021_01001";
%%if you load your own data in other format, you can comment out next
%%line.
%we use 4 points to generate a 3D convex hull as the Lev_ini.
[Img,Lev_ini,ground_truth]=get_image;
%% If ground_truth is not usually available, let it be Img

disp('  Img, Lev_ini, ground_truth  are  ready')
scr = get(0,'ScreenSize');
figure('Position',[10 10 scr(3:4)/2])
my_draw(Lev_ini, Img,Lev_ini,'Initial markers'); drawnow
title('Initial Set up')

figure('Position',[scr(3:4)/3 scr(3:4)/1.8])
%% Data Img,Lvset_ini,Label_gt are ready
disp('  Compute Ricci term rcc')
rcc=getrcc(Img);
rcc=1./(1+abs(rcc));
[seg,u] = level_set(Img, rcc,Lev_ini,[0.32	1	1.1	200	-2	140] );
 
dice_metric=dice(seg,ground_truth==1);

format compact
dice_metric, my_draw(seg,Img,u, 'Final Seg');


%%
function [Img,Lev_ini,ground_truth]=get_image
%this function can read your data, e.g., 4 channel 3D image in this image.
path="BraTS2021_01001";
Img=getimage(path);
ground_truth=niftiread(fullfile(path,ls(fullfile(path,"*seg.nii.gz"))));

Lev_ini=zeros(size(Img,[2,3,4]),"logical");
Lev_ini(133,123,74)=1;
Lev_ini(130,126,65)=1;
Lev_ini(136,122,73)=1;
Lev_ini(132,126,73)=1;
stats=regionprops3(uint8(Lev_ini),"ConvexImage","SubarrayIdx");
Lev_ini(stats.SubarrayIdx{:})=stats.ConvexImage{1};
end

function image=RescaleIntensity(image)
min_num=min(image(:));
acc=accumarray(image(:)-min_num+1,ones(numel(image),1))/numel(image);
acc=cumsum(acc);
id=find(acc>0.995,1);
image(image>min_num-1+id)=min_num-1+id;
image=mat2gray(image);
end

function Img=getimage(path)
Img1=niftiread(fullfile(path,ls(fullfile(path,"*t1ce.nii.gz"))));
Img2=niftiread(fullfile(path,ls(fullfile(path,"*flair.nii.gz"))));
Img3=niftiread(fullfile(path,ls(fullfile(path,"*t1.nii.gz"))));
Img4=niftiread(fullfile(path,ls(fullfile(path,"*t2.nii.gz"))));
Img=cat(4,RescaleIntensity(Img1),RescaleIntensity(Img2),RescaleIntensity(Img3),RescaleIntensity(Img4));
Img=permute(Img,[4,1,2,3]);
Img=single(Img);

end

function rcc=getrcc(Img)
ss=size(Img);
rcc=zeros(ss(2:end));
for i=1:size(Img,1)
rcc=rcc+(get_ricci(imgaussfilt3(squeeze(Img(i,:,:,:)),2) ,[1,1,1]));
end

end
function [bw,u] = level_set(Img,rcc,bw ,par )
disp('   Method starts')
c1=2;
c2=-2;
u = (bwdist(1 - bw) - bwdist(bw)...
    - (bw - 0.5));
u(u>c1)=c1;
u(u<c2)=c2;

u=level_set_4_files_parallel(padarray(single(u),[1,1,1],-2,'both'), ...
    padarray(single(Img),[0,1,1,1],"replicate","both"),  ...
    padarray(single(rcc),[1,1,1],"replicate","both"), ...
    padarray(uint8(bw),[1,1,1],3,"both"),par);
u=u(2:end-1,2:end-1,2:end-1);
    bw=u>0;%bw=u(2:end-1,2:end-1,2:end-1)>0;
end
function my_draw(seg,Img, u,txt)
    I=imgaussfilt3(single(seg),0.3);     % my_draw(seg);
    
    subplot(221)
    ax1=Adraw3D(I); colormap(ax1,hsv); 
    a1=55; a2=250;  view(a1,a2); title(txt)
    
    ax2=subplot(222);
    imagesc( squeeze(Img(1,132,:,:)) ); axis tight; colormap(ax2,gray); hold on
    contour(squeeze(u(132,:,:)),[0.9,1],'r','linewidth',3);
    ax2=gca; title('x slice 132'); view(0,90)

    ax3=subplot(223);
    imagesc( squeeze(Img(1,:,133,:)) ); axis tight; colormap(ax3,gray); hold on
    contour(squeeze(u(:,124,:)),[0.9,1],'r','linewidth',3);
    ax3=gca; title('y slice 133')
    
    ax4=subplot(224);
    imagesc( squeeze(Img(1,:,:,73)) ); axis tight; colormap(ax4,gray); hold on
    contour(squeeze(u(:,:,73)),[0.9,1],'r','linewidth',3);
    ax4=gca; title('z slice 73')
    
    %hlink = linkprop([ax1,ax2,ax3,ax4],{'CameraPosition','CameraUpVector','PlotBoxAspectRatio'});
end

function draw(img,seg)

u=imgaussfilt3(single(seg),0.3);

nexttile
Adraw3D(u);

nexttile
img_slice=squeeze(img(124,:,:));
label_slice=squeeze(seg(124,:,:));
img_x=imoverlay(img_slice,boundarymask(label_slice));
imshow(img_x);


nexttile
img_slice=squeeze(img(:,129,:));
label_slice=squeeze(seg(:,129,:));
img_y=imoverlay(img_slice,boundarymask(label_slice));
imshow(img_y);


nexttile
img_slice=squeeze(img(:,:,73));
label_slice=squeeze(seg(:,:,73));
img_z=imoverlay(img_slice,boundarymask(label_slice));
imshow(img_z);

end

function ax=draw3D(u)
    fv = isosurface(u,0.9);
    V=fv.vertices;
    VV=vertexNormal(triangulation(fv.faces,V));
    pa=patch('Faces',fv.faces,'Vertices',V,'FaceVertexCData',V(:,3),'FaceColor','interp','EdgeColor','none','VertexNormals',VV,'FaceLighting','gouraud');
    pa.Clipping='off';
    ax=gca;
    material shiny;
    light
    axis tight manual equal off;
end


function ax=Adraw3D(u)

fv = isosurface(u,0.9);
V=fv.vertices;
VV=vertexNormal(triangulation(fv.faces,V));
pa=patch('Faces',fv.faces,'Vertices',V,'FaceVertexCData',V(:,3),'FaceColor','interp','EdgeColor','none','VertexNormals',VV,'FaceLighting','gouraud');
pa.Clipping='off'; ax=gca;
material shiny;
light
axis tight manual equal off;
end