

#include <iostream>
#include <vector>
#include <cstdlib>
#include <cmath>
#include <array>
#include <cassert>
#include <functional>
#include <mex.h> 
#include "matrix.h"

class MyClass
{
public:
	MyClass(mxArray* plhs[], const mxArray* prhs[]);
	~MyClass();
	void step2();
	void step2_1();
	void step3();
	void step3_1();
private:
	const double* weight;
	double* nweight;
	mwSize ni, nj, nk;
	const mwSize* dims;
};

MyClass::MyClass(mxArray* plhs[], const mxArray* prhs[])
{


	dims = mxGetDimensions(prhs[0]);
	plhs[0] = mxCreateNumericArray(mxGetNumberOfDimensions(prhs[0]), dims, mxDOUBLE_CLASS, mxREAL);
	nweight = (double*)mxGetPr(plhs[0]);
	ni = dims[0];
	nj = dims[1];
	nk = dims[2];
	weight = (double*)mxGetPr(prhs[0]);

	void (MyClass:: * p[])() = { &MyClass::step2_1,&MyClass::step3,&MyClass::step3_1 };
	(this->*p[int(mxGetScalar(prhs[1]))])();
	/*step3();*/


};

MyClass::~MyClass()
{
};

void MyClass::step2()
{
	double d1, d2, d3, d4;
	int i1, i2, i3, i4;
	std::array < mwSize, 3 >offset = { 1, ni, ni * nj };
	mwSize s1, s2, s3, li, lj, lk;
	std::array<mwSize, 3> temp;
	auto intersection = [&s1, &d2, &offset, this](int op1, int op2)-> double
	{
		double d = sqrt(weight[s1 + offset[op1] + offset[op2]] * d2);
		double rd = abs(weight[s1 + offset[op2]] / d - d / weight[s1 + offset[op1]]);
		d = sqrt(weight[s1 - offset[op1] + offset[op2]] * d2);
		rd += abs(weight[s1 + offset[op2]] / d - d / weight[s1 - offset[op1]]);
		d = sqrt(weight[s1 - offset[op1] - offset[op2]] * d2);
		rd += abs(weight[s1 - offset[op2]] / d - d / weight[s1 - offset[op1]]);
		d = sqrt(weight[s1 + offset[op1] - offset[op2]] * d2);
		rd += abs(weight[s1 - offset[op2]] / d - d / weight[s1 + offset[op1]]);
		return rd;


	};
	auto add1 = [&s1, &d2, &offset, this](int op1, int op2)-> double
	{


		return (weight[s1 + offset[op1]] + weight[s1 - offset[op1]] + weight[s1 - offset[op2]] + weight[s1 + offset[op2]]) / d2;


	};
	auto subtract1 = [&s1, &d2, &offset, &li, &lj, &lk, &temp, this](int op1)-> double
	{


		double d = 0;
		temp = { li,lj,lk };
		if (temp[op1] > 1)
			d += weight[s1 - offset[op1]] / sqrt(weight[s1 - offset[op1] - offset[op1]] * d2);
		if (temp[op1] < dims[op1] - 2)
			d += weight[s1 + offset[op1]] / sqrt(weight[s1 + offset[op1] + offset[op1]] * d2);
		return  d;


	};

	auto add2 = [&s1, &d2, &offset, this](int op1)-> double
	{


		return d2 / (weight[s1 + offset[op1]]) + d2 / (weight[s1 - offset[op1]]);


	};
	auto subtract2 = [&s1, &d2, &offset, this](int op1)-> double
	{


		return sqrt(weight[s1 - offset[op1] - offset[op1]] * d2) / weight[s1 - offset[op1]] + sqrt(weight[s1 + offset[op1] + offset[op1]] * d2) / weight[s1 + offset[op1]];



	};
	for (li = 2; li < ni - 2; li += 2)
		for (lj = 1; lj < nj - 1; lj += 2)
			for (lk = 1; lk < nk - 1; lk += 2)
			{
				s1 = lk * offset[2] + lj * offset[1] + li;
				d2 = weight[s1];
				d1 = 0;
				d1 += add1(1, 2);
				d1 += add2(0);
				//d1 -= intersection(0, 1);
				//d1 -= intersection(0, 2);
				assert(abs(d1) < 10000);
				d1 -= subtract1(1);
				assert(abs(d1) < 10000);
				d1 -= subtract1(2);
				assert(abs(d1) < 10000);
				d1 -= subtract2(0);

				assert(abs(d1) < 10000);
				nweight[s1] = d1;

			}
	for (li = 1; li < ni - 1; li += 2)
		for (lj = 2; lj < nj - 2; lj += 2)
			for (lk = 1; lk < nk - 1; lk += 2)
			{
				s1 = lk * offset[2] + lj * offset[1] + li;
				d2 = weight[s1];
				d1 = 0;
				d1 += add1(0, 2);
				d1 += add2(1);
				//d1 -= intersection(1, 0);
				//d1 -= intersection(1, 2);
				d1 -= subtract1(0);
				d1 -= subtract1(2);
				d1 -= subtract2(1);
				assert(abs(d1) < 10000);
				nweight[s1] = d1;

			}
	for (li = 1; li < ni - 1; li += 2)
		for (lj = 1; lj < nj - 1; lj += 2)
			for (lk = 2; lk < nk - 2; lk += 2)
			{
				s1 = lk * offset[2] + lj * offset[1] + li;
				d2 = weight[s1];
				d1 = 0;
				d1 += add1(0, 1);
				d1 += add2(2);
				//d1 -= intersection(2, 0);
				//d1 -= intersection(2, 1);
				d1 -= subtract1(0);
				d1 -= subtract1(1);
				d1 -= subtract2(2);
				assert(abs(d1) < 10000);
				nweight[s1] = d1;

			}
}
void MyClass::step2_1()
{
	double d1, d2, d3, d4;
	int i1, i2, i3, i4;
	std::array < mwSize, 3 >offset = { 1, ni, ni * nj };
	mwSize s1, s2, s3, li, lj, lk;
	std::array<mwSize, 3> temp;
	double* p = nweight;
	for (li = 0, lj = ni * nj * nk; li < lj; li++, p++)
		*(p) = std::nan("");
	auto div_1 = [](double a, double b)-> double {
		return a / b;
	};
	auto div_2 = [](double a, double b)-> double {
		return b / a;
	};
	std::array<std::function<double(double, double)>, 2> div = { div_1,div_2 };

	auto add = [&s1, &d1, &d2, &d3, &offset, &temp, &div, this](int op, int divop)-> double
	{

		double d = 0;
		if (temp[op] > 1) {
			d1 = weight[s1 - 2 * offset[op]];
			d3 = weight[s1 - offset[op]];
			d += (div[divop](d2, d3) - div[divop](sqrt(d2* d1), d3));
		}
		if (temp[op] + 2 < dims[op]) {
			d1 = weight[s1 + 2 * offset[op]];
			d3 = weight[s1 + offset[op]];
			d += (div[divop](d2, d3) - div[divop](sqrt(d2* d1), d3));
		}
		return  d;
	};
	auto get_offset = [&temp, &offset]()->mwSize {
		return temp[2] * offset[2] + temp[1] * offset[1] + temp[0];
	};
	for (li = 1; li < ni; li += 2)
		for (lj = 1; lj < nj; lj += 2)
			for (lk = 1; lk < nk; lk += 2) {
				temp = { li,lj,lk };
				s1 = get_offset();
				for (i1 = 0; i1 < 3; i1++) {
					temp[i1] += 1;
					s1 += offset[i1];
					d2 = weight[s1];
					{
						double& tw = nweight[s1];
						tw = std::isnan(tw) ? add(i1, 0) + add((i1 + 1) % 3, 1) + add((i1 + 2) % 3, 1) : tw;
					}



					temp[i1] -= 2;
					s1 -= 2 * offset[i1];
					d2 = weight[s1];
					{
						double& tw = nweight[s1];
						tw = std::isnan(tw) ? add(i1, 0) + add((i1 + 1) % 3, 1) + add((i1 + 2) % 3, 1) : tw;
					}


					temp[i1] += 1;
					s1 += offset[i1];

				}


			}



}
void MyClass::step3()
{
	double d1, d2, d3;
	mwSize s1, s2, s3;
	std::array<mwSize, 3> temp;

	std::array < mwSize, 3 >offset = { 1, ni, ni * nj };
	for (mwSize li = 1; li < ni; li += 2)
		for (mwSize lj = 1; lj < nj; lj += 2)
			for (mwSize lk = 1; lk < nk; lk += 2) {
				s1 = lk * offset[2] + lj * offset[1] + li;
				d1 = 0;
				d2 = weight[s1];
				temp = { li,lj,lk };
				for (mwSize kk = 0; kk < 3; kk++) {


					if (temp[kk] + 2 < dims[kk]) {
						s2 = s1 + 2 * offset[kk];
						d3 = weight[s1 + offset[kk]];
						d1 += (d3 / d2 - d3 / sqrt(d2 * weight[s2]));
					}

					if (temp[kk] > 1) {
						s2 = s1 - 2 * offset[kk];
						d3 = weight[s1 - offset[kk]];
						d1 += (d3 / d2 - d3 / sqrt(d2 * weight[s2]));
					}
				}
				nweight[s1] = d1;

			}
}

void MyClass::step3_1()
{
	double d1, d2, d3;
	mwSize s1, s2, s3;
	std::array<mwSize, 3> temp;

	std::array < mwSize, 3 >offset = { 1, ni, ni * nj };
	for (mwSize li = 1; li < ni; li += 2)
		for (mwSize lj = 1; lj < nj; lj += 2)
			for (mwSize lk = 1; lk < nk; lk += 2) {
				s1 = lk * offset[2] + lj * offset[1] + li;
				d1 = 0;
				d2 = weight[s1];
				temp = { li,lj,lk };
				for (mwSize kk = 0; kk < 3; kk++) {


					if (temp[kk] + 2 < dims[kk]) {
						s2 = s1 + 2 * offset[kk];
						d3 = weight[s1 + offset[kk]];
						d1 += (d2 / d3 - sqrt(d2 * weight[s2]) / d3);
					}

					if (temp[kk] > 1) {
						s2 = s1 - 2 * offset[kk];
						d3 = weight[s1 - offset[kk]];
						d1 += (d2 / d3 - sqrt(d2 * weight[s2]) / d3);
					}
				}
				nweight[s1] = d1;

			}
}


/* MEX entry function */
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])

{
	MyClass ff(plhs, prhs);


}
